package bigal 
{
	import bigal.utils.AgalBlendMode;
	import bigal.utils.AgalUtils;
	import flash.display.BitmapData;
	import flash.display3D.Context3D;
	import flash.display3D.Context3DProgramType;
	import flash.display3D.Context3DTextureFormat;
	import flash.display3D.Context3DVertexBufferFormat;
	import flash.display3D.IndexBuffer3D;
	import flash.display3D.Program3D;
	import flash.display3D.textures.Texture;
	import flash.display3D.VertexBuffer3D;
	/**
	 * @author Nicolas Barradeau
	 * http://en.nicoptere.net
	 */
	public class Model extends Object3D
	{
		//coordinates
		private var _vertices:Vector.<Number>;
		private var _indices:Vector.<uint>;
		private var _colors:Vector.<Number>;
		private var _normals:Vector.<Number>;
		private var _uvs:Vector.<Number>;
		private var _bitmapData:BitmapData;
		
		//AGAL props
		protected var vertexBuffer:VertexBuffer3D;
		protected var colorBuffer:VertexBuffer3D;
		protected var normalBuffer:VertexBuffer3D;
		protected var uvBuffer:VertexBuffer3D;
		protected var indexBuffer:IndexBuffer3D;
		protected var texture:Texture;
		
		
		protected var _vertexShader:String =	"m44 op, va0, vc0\n"	+
												"mov v0, va0";
		protected var _fragmentShader:String = 	"mov oc, v0\n";
		
		private var _program:Program3D;
		private var programDirty:Boolean = true;
		
		private var _blendMode:String = AgalBlendMode.NORMAL;
		private var _visible:Boolean = true;
		
		public function Model( vertices:Vector.<Number> = null, indices:Vector.<uint> = null ) 
		{
			
			this.vertices = vertices;
			this.indices = indices;
			
		}
		
		
		/**
		 * prepares the context3D to render. by default it assigns the vertexBuffer as va0
		 * @param	context the Context3D to target
		 */
		public function prepareContext( context:Context3D ):void
		{
			
			AgalBase.setVertexBufferAt( 0, vertexBuffer, 0, Context3DVertexBufferFormat.FLOAT_3 );
			if ( blendMode != AgalBlendMode.NORMAL ) AgalBlendMode.apply( AgalBase.context, blendMode );
			
		}
		/**
		 * render pass
		 * @param	context
		 */
		public function render( context:Context3D ):void
		{
			
			prepareContext( context );
			
			context.setProgram( program );
			context.drawTriangles( indexBuffer ); 
			
			clearContext( context );
			
		}
		
		/**
		 * clears all the resources used during the render. by default, it unassigns the vertexBuffer as va0
		 * @param	context the Context3D to clear
		 */
		public function clearContext( context:Context3D ):void
		{
			
			AgalBase.setVertexBufferAt( 0, null );
			AgalBlendMode.apply( AgalBase.context, AgalBlendMode.NORMAL );
			
		}
		
		/****************************************
		
			GEOMETRY
		
		****************************************/
		
		public function get vertices():Vector.<Number> {	return _vertices;	}
		public function set vertices(value:Vector.<Number>):void 
		{
			_vertices = popNaN( value );
			if ( vertexBuffer != null ) vertexBuffer.dispose();
			if ( _vertices == null ) return;
			vertexBuffer = AgalBase.getVertexBuffer( _vertices, 3 );
		}
		
		
		public function get indices():Vector.<uint> {	return _indices;	}
		public function set indices(value:Vector.<uint>):void 
		{
			_indices = popUint( value );
			if ( indexBuffer != null ) indexBuffer.dispose();
			if ( _indices == null ) return;
			indexBuffer = AgalBase.getIndexBuffer( _indices );
		}
		
		
		public function get colors():Vector.<Number> {		return _colors;	}
		public function set colors(value:Vector.<Number>):void 
		{
			_colors = popNaN( value );
			if ( colorBuffer != null ) colorBuffer.dispose();
			if ( _colors == null ) return;
			colorBuffer = AgalBase.getVertexBuffer( _colors, 3 );
		}
		
		
		public function get uvs():Vector.<Number> {		return _uvs;	}
		public function set uvs(value:Vector.<Number>):void 
		{
			_uvs = popNaN( value );
			if ( uvs == null ) return;
			if ( uvBuffer != null ) uvBuffer.dispose();
			uvBuffer = AgalBase.getVertexBuffer( _uvs, 2 );
		}
		
		
		public function get normals():Vector.<Number> {		return _normals;	}
		public function set normals(value:Vector.<Number>):void 
		{
			_normals = popNaN( value );
			if ( normalBuffer != null ) normalBuffer.dispose();
			if ( _normals == null ) return;
			normalBuffer = AgalBase.getVertexBuffer( _normals, 3 );
		}
		
		
		public function get bitmapData():BitmapData {	return _bitmapData;		}
		public function set bitmapData(value:BitmapData):void 
		{
			
			_bitmapData = AgalUtils.validateBitmapData( value );
			if ( _bitmapData == null ) return;
			texture = AgalBase.getTexture( _bitmapData, Context3DTextureFormat.BGRA, false );
			
		}
		
		public function get blendMode():String {	return _blendMode;	}
		public function set blendMode(value:String):void 
		{
			_blendMode = value;
		}
		
		public function get visible():Boolean {	return _visible;	}
		public function set visible(value:Boolean):void 
		{
			_visible = value;
		}
		
		
		
		
		/****************************************
		
			SHADER
		
		****************************************/
		
		public function get vertexShader():String 
		{
			return _vertexShader;
		}
		
		public function set vertexShader(value:String):void 
		{
			_vertexShader = value;
			programDirty = true;
		}
		
		public function get fragmentShader():String 
		{
			return _fragmentShader;
		}
		
		public function set fragmentShader(value:String):void 
		{
			_fragmentShader = value;
			programDirty = true;
		}
		/**
		 * provides the program3D
		 */
		public function get program():Program3D 
		{
			if ( programDirty ) 
			{
				resetProgram();
				programDirty = false;
			}
			return _program;
		}
		
		protected function resetProgram():void 
		{
			
			_program ||= AgalBase.context.createProgram();
			
			_program.upload(	AgalBase.assembler.assemble(Context3DProgramType.VERTEX, _vertexShader )	,
								AgalBase.assembler.assemble(Context3DProgramType.FRAGMENT, _fragmentShader 	)	); 
			
		}
		
		//make sure there's no empty slot at the end of the vectors
		private function popUint( vector:Vector.<uint> = null ):Vector.<uint>
		{
			if ( vector == null ) return null;
			var i:int = vector.length - 1;
			while ( isNaN( vector[ i-- ] ) ) 
			{
				vector.pop(); 
			}
			return vector;
		}
		private function popNaN( vector:Vector.<Number> = null ):Vector.<Number>
		{
			if ( vector == null ) return null;
			var i:int = vector.length - 1;
			while ( isNaN( vector[ i-- ] ) ) 
			{
				vector.pop(); 
			}
			return vector;
		}
		
		
	}

}