package bigal.utils 
{
	import flash.display.BitmapData;
	import flash.display3D.Context3D;
	import flash.display3D.Context3DTextureFormat;
	import flash.display3D.textures.CubeTexture;
	import flash.geom.Matrix;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	/**
	 * @author Nicolas Barradeau
	 * http://en.nicoptere.net
	 */
	public class AgalUtils 
	{
		
		/*********************************************
		
		VECTOR.<NUMBER> UTILS
		
		*********************************************/
		
		/**
		 * shrinks a Vector.<Number> down to a given length
		 * @param	values original vector of values
		 * @param	length the desired length
		 * @return a vector of values matching the desired length
		 */
		static public function compress( values:Vector.<Number>, length:uint ):Vector.<Number>
		{
			var tmp:Vector.<Number> = new Vector.<Number>();
			var i:Number;
			if ( values.length == 1 )
			{
				i = length;
				while ( i-- ) tmp.push( values[0] );
				return tmp;
			}
			
			var inputLength:uint = values.length - 2;
			var outputLength:uint = length - 2;
			tmp.push( values[ 0 ] );
			i = 0;
			var j:int = 0;
			while (j < inputLength) 
			{
				var diff:uint = ( i + 1 ) * inputLength - ( j + 1 ) * outputLength;
				if (diff < inputLength / 2 ) 
				{
					i++;
					tmp.push( values[ j++ ] );
				}
				else j++;
			}
			tmp.push( values[ inputLength + 1 ] );
			return tmp;
		}
		
		/**
		 * compresses 2 vectors of numbers
		 * @param	u first set of vectors to compress
		 * @param	v second set of vectors to compress
		 * @param	length0 the desired length for vector u
		 * @param	length1 the desired length for vector v
		 * @return a Vector.<Number> of doublets (x,y) of the given length interpolating the original values
		 */
		static public function compress2D( u:Vector.<Number>, v:Vector.<Number>, length0:uint, length1:uint ):Vector.<Number>
		{
			
			var _u:Vector.<Number> = compress( u, length0 );
			var _v:Vector.<Number> = compress( v, length1 );
			var tmp:Vector.<Number> = new Vector.<Number>();
			
			for (var i:int = 0; i < length0; i++) 
			{
				for (var j:int = 0; j < length1; j++) 
				{
					
					tmp.push( _u[ i ], _v[ j ] );
					
				}
			}
			
			return tmp;
			
		}
		
		/**
		 * compresses 3 vectors of numbers
		 * @param	u first set of vectors to compress
		 * @param	v second set of vectors to compress
		 * @param	w second set of vectors to compress
		 * @param	length0 the desired length for vector u
		 * @param	length1 the desired length for vector v
		 * @param	length2 the desired length for vector w
		 * @return a Vector.<Number> of triplets (x,y,z) of the given length interpolating the original values
		 */
		static public function compress3D( u:Vector.<Number>, v:Vector.<Number>, w:Vector.<Number>, length0:uint, length1:uint, length2:uint ):Vector.<Number>
		{
			
			var _u:Vector.<Number> = compress( u, length0 );
			var _v:Vector.<Number> = compress( v, length1 );
			var _w:Vector.<Number> = compress( w, length2 );
			
			var tmp:Vector.<Number> = new Vector.<Number>();
			
			for (var i:int = 0; i < length0; i++) 
			{
				for (var j:int = 0; j < length1; j++) 
				{
					for (var k:int = 0; k < length2; k++) 
					{
						
						tmp.push( _u[ i ], _v[ j ], _w[ k ] );
						
					}
				}
			}
			
			return tmp;
			
		}
		
		/**
		 * expands the content of a vector.<Number> to match a given length
		 * @param	values original vector of values
		 * @param	length the desired length of the new Vector
		 * @return a Vector.<Number> of the given length interpolating the original values
		 */
		static public function expand( values:Vector.<Number>, length:uint ):Vector.<Number>
		{
			var tmp:Vector.<Number> = new Vector.<Number>();
			var i:Number, j:Number;
			if ( values.length == 1 )
			{
				i = length;
				while ( i-- ) tmp.push( values[0] );
				return tmp;
			}
			
			var step:Number = 1 / ( ( values.length - 1 ) / (length - 1) );
			var val0:Number, val1:Number, max:Number, delta:int;
			for ( i = 0; i < length; i+= step ) 
			{
				val0 = values[ int( i / step ) ];
				
				max = ( int( i / step ) + 1 );
				if ( max > values.length - 1 ) max = values.length - 1;
				
				val1 = values[ max ];
				
				delta = ( i + step > length ) ? ( length - i ) : step;
				
				for ( j = 0; j < delta; j++) 
				{ 
					tmp.push( lerp( j / delta, val0, val1 ) );
				}
				
			}
			while ( tmp.length < length ) tmp.push( values[ values.length - 1 ] );
			
			return tmp;
			
		}
		
		/**
		 * expands 2 vectors of numbers and combines them together
		 * @param	u first set of vectors to expand
		 * @param	v second set of vectors to expand
		 * @param	length0 the desired length for vector u
		 * @param	length1 the desired length for vector v
		 * @return a Vector.<Number> of doublets (x,y) of the given length interpolating the original values
		 */
		static public function expand2D( u:Vector.<Number>, v:Vector.<Number>, length0:uint, length1:uint ):Vector.<Number>
		{
			
			var _u:Vector.<Number> = expand( u, length0 );
			var _v:Vector.<Number> = expand( v, length1 );
			var tmp:Vector.<Number> = new Vector.<Number>();
			
			for (var i:int = 0; i < length0; i++) 
			{
				for (var j:int = 0; j < length1; j++) 
				{
					
					tmp.push( _u[ i ], _v[ j ] );
					
				}
			}
			return tmp;
			
		}
		
		/**
		 * expands 3 vectors of numbers
		 * @param	u first set of vectors to expand
		 * @param	v second set of vectors to expand
		 * @param	w second set of vectors to expand
		 * @param	length0 the desired length for vector u
		 * @param	length1 the desired length for vector v
		 * @param	length2 the desired length for vector w
		 * @return a Vector.<Number> of triplets (x,y,z) of the given length interpolating the original values
		 */
		static public function expand3D( u:Vector.<Number>, v:Vector.<Number>, w:Vector.<Number>, length0:uint, length1:uint, length2:uint ):Vector.<Number>
		{
			
			var _u:Vector.<Number> = expand( u, length0 );
			var _v:Vector.<Number> = expand( v, length1 );
			var _w:Vector.<Number> = expand( w, length2 );
			
			var tmp:Vector.<Number> = new Vector.<Number>();
			
			for (var i:int = 0; i < length0; i++) 
			{
				for (var j:int = 0; j < length1; j++) 
				{
					for (var k:int = 0; k < length2; k++) 
					{
						
						tmp.push( _u[ i ], _v[ j ], _w[ k ] );
						
					}
				}
			}
			return tmp;
			
		}
		
		/*********************************************
		
		BITMAPDATA UTILS
		
		*********************************************/
		
		/**
		 * make sure that a given bitmapData's dimensions will be a power of 2
		 * @param	bitmapData source bitmapdata
		 * @param	stretch	specifies if the new bitmapData should be stretched
		 * @return
		 */
		static public function validateBitmapData( bitmapData:BitmapData, stretch:Boolean = true ):BitmapData
		{
			var w:uint = nextPowerOfTwo( bitmapData.width );
			var h:uint = nextPowerOfTwo( bitmapData.height );
			
			//if the bitmap is valide, do nothing
			if ( w == bitmapData.width && h == bitmapData.height ) return bitmapData;
			
			//otherwise creates a power of 2 bitmap and draws the original inside
			var tmp:BitmapData = new BitmapData( w, h, bitmapData.transparent, 0 );
			var m:Matrix = new Matrix();
			if ( stretch )
			{
				m.scale( w / bitmapData.width, h / bitmapData.height );
			}
			tmp.draw( bitmapData, m );
			bitmapData = tmp;
			return bitmapData;
			
		}
		
		static public function nextPowerOfTwo( value:Number ):uint
		{
			if( value < 2 || value > 2048 ) return 2;
			var c : uint = 1;
			while( c < value ) c = c << 1;
			return c;
		}
		
		static public function createCubeMap( context:Context3D, bd:BitmapData, method:int = 0 ):CubeTexture
		{
			
			var w:int = bd.width;
			var h:int = bd.height;
			
			var step_x:int = 4;
			var step_y:int = 3;
			
			var order:Array = [];
			switch( method )
			{
				case 0:
					
					/**
					 * __|y+|__|__
					 * x-|z-|x+|z+
					 * __|y-|__|__
					 */
					
					order = [ 	-1,  2, -1, -1,
								 1,  4,  0,  5,
								-1,  3, -1, -1 	];
					break;
					
				case 1:
					
					/**
					 * __|y+|__|__
					 * x-|z-|x+|z+
					 * __|__|y-|__
					 */
					
					order = [ 	-1,  2, -1, -1,
								 1,  4,  0,  5,
								-1, -1,  3, -1 	];

					break;
					
				case 2:
					
					/**
					 * __|y+|__
					 * x-|z-|x+
					 * __|y-|__
					 * __|z+|__
					 */
					step_x = 3;
					step_y = 4;
					order = [ 	-1,  2, -1, 
								 1,	 4,  0,
								 -1, 3,	-1,
								 -1, 5, -1	];
					break;
			}
			
			var rect:Rectangle = new Rectangle( 0, 0, int( w / step_x ), int( h / step_y ) );
			
			var tmp:BitmapData = new BitmapData( int( w / step_x ), int( h / step_y ), true, 0xFF0000);
			
			var ct:CubeTexture = context.createCubeTexture( rect.width, Context3DTextureFormat.BGRA, false );
			
			for ( var i:int = 0; i < order.length; i++ )
			{
				if ( order[ i ] != -1 )
				{
					
					rect.x = i % step_x * rect.width; 
					rect.y = int( i / step_x ) * rect.height; 
					
					tmp.fillRect( tmp.rect, 0xFF0000 );
					
					tmp.copyPixels( bd, rect, new Point(), null, null, true );
					
					ct.uploadFromBitmapData( tmp.clone(), order[ i ] );
					
				}
				
			}
			
			return ct;
			
		}
		
		/*********************************************
		
		ALL TIME UTILS
		
		*********************************************/
		
		static public function normalize(value:Number, minimum:Number, maximum:Number):Number
		{
			return (value - minimum) / (maximum - minimum);
		}
		static public function lerp(normValue:Number, minimum:Number, maximum:Number):Number
		{
			return minimum + (maximum - minimum) * normValue;
		}
		static public function map(value:Number, min1:Number, max1:Number, min2:Number, max2:Number):Number
		{
			return lerp( normalize(value, min1, max1), min2, max2);
		}

	}

}